from pathlib import Path
from typing import List

prompts = {
    path.stem: open(path).read().strip() for path in Path(__file__).parent.glob("*.txt")
}


def step_by_step(prompt: str) -> str:
    return prompts["step_by_step"].format(prompt=prompt)


def you_are_a_researcher(question: str, candidates: List[str]) -> str:
    answers = []
    for idx, candidate in enumerate(candidates):
        answers.append(f"Answer option {idx+1}:\n\n{candidate}")

    return prompts["researcher"].format(
        N=len(candidates),
        question=question,
        answer_list="\n\n".join(answers),
    )


def you_are_a_resolver(candidates: List[str], silence_rationale: bool) -> str:
    exclude_rational = (
        "Please do steps 1) and 2) silently and only print the improved answer. "
        "Exclude mention that it is an improved answer."
    )

    return prompts["resolver"].format(
        N=len(candidates),
        exclude_rationale=exclude_rational if silence_rationale else "",
    )



#! /usr/bin/env python

import sys
import textwrap

from smartgpt.util import color_text


def remove_spaces(text):
    if not text:
        return ""

    while True:
        if text.find("  ") > -1:
            text = text.replace("  ", " ")
        else:
            break

    return text


class SmartGPTError(Exception):
    def __init__(self, e=None):
        self.e = remove_spaces(e)
        self.error_type = "SmartGPT Error"
        Exception.__init__(self)
        return

    def __str__(self):
        max_len = max([len(line) for line in textwrap.fill(self.e, 80).split("\n")])
        error_lines = [
            "%s%s" % (line, " " * (max_len - len(line)))
            for line in textwrap.fill(self.e, 80).split("\n")
        ]

        error_message = [
            "%s: %s" % (color_text(self.error_type, "red"), error_lines[0])
        ]
        for error_line in error_lines[1:]:
            error_message.append(
                "%s%s" % (" " * (len(self.error_type) + 2), error_line)
            )

        return "\n\n" + "\n".join(error_message) + "\n\n"

    def clear_text(self):
        return self.e



import smartgpt.user_profile
from smartgpt.chat import Agent, SmartGPT
from smartgpt.datatypes import (
    Credentials,
    GPTConfig,
    Message,
    Mode,
    Response,
    Role,
    Settings,
    Verbosity,
)

__version__ = "0.1.0"



import logging
import sys

from smartgpt.datatypes import Verbosity


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors"""

    blue = "\x1b[34;1m"
    yellow = "\x1b[33;1m"
    green = "\x1b[32;1m"
    reset = "\x1b[0m"
    format_str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    def format(self, record):
        record.levelname = self.yellow + record.levelname + self.reset
        record.name = self.blue + record.name + self.reset
        return super().format(record)

    def formatMessage(self, record):
        # Override asctime in the format string with the colorized version
        return self.format_str % {
            "asctime": self.green + self.formatTime(record, self.datefmt) + self.reset,
            "name": record.name,
            "levelname": record.levelname,
            "message": record.getMessage(),
        }


handler = logging.StreamHandler(sys.stdout)
handler.setFormatter(CustomFormatter())


def get_logger(
    name: str = "SmartGPT", verbosity: Verbosity = Verbosity.SOME
) -> logging.Logger:
    logger = logging.getLogger(name=name)
    logger.addHandler(handler)
    logger.propagate = False

    if verbosity == Verbosity.NONE:
        logger.setLevel(logging.CRITICAL)
    elif verbosity == Verbosity.SOME:
        logger.setLevel(logging.INFO)
    elif verbosity == Verbosity.ALL:
        logger.setLevel(logging.DEBUG)
    else:
        raise NotImplementedError()

    return logger


default_logger = get_logger()



"""
This module defines fundamental data types and configurations for interacting with
the GPT model. It provides enums for interaction modes and roles, structures for
messages, responses and configurations.

Classes:
    Mode:
        Enum representing the different modes of interaction.
    GPTConfig:
        Configuration for the GPT model.
    Role:
        Enum representing the different roles in a conversation.
    Message:
        Structure for encapsulating a message in a conversation.
    Response:
        Structure for encapsulating a response from the GPT model.
"""

from __future__ import annotations

from pathlib import Path
from typing import List

import attrs
import cattrs
import yaml

from smartgpt import strenum
from smartgpt.strenum import StrEnum
from smartgpt.util import Pathish

cattrs.register_unstructure_hook(
    StrEnum,
    lambda v: v.value,
)


@attrs.define
class Settings:
    generator_temps: List[float]
    researcher_temp: float
    resolver_temp: float
    model: str
    mode: Mode
    credentials: Credentials
    verbosity: Verbosity
    vi_mode: bool

    def __repr__(self):
        margin = " " * 4
        lines = [f"\n{margin}Runtime Settings"]
        middle_prefix = f"{margin}├── "
        final_prefix = f"{margin}└── "

        attr_names = [attr.name for attr in attrs.fields(Settings)]
        max_len = max(len(attr_name) for attr_name in attr_names)

        for idx, attr_name in enumerate(attr_names):
            attr_val = getattr(self, attr_name)
            spacer = " " * (max_len - len(attr_name))
            prefix = final_prefix if idx == len(attr_names) - 1 else middle_prefix
            lines.append(f"{prefix}{attr_name}{spacer} = {attr_val.__str__()}")

        return "\n".join(lines) + "\n"

    @property
    def num_agents(self) -> int:
        return len(self.generator_temps)

    def save(self, path: Pathish) -> Path:
        path = Path(path)
        with path.open("w") as file:
            yaml.dump(cattrs.unstructure(self), file)

        return path

    @classmethod
    def load(cls, path: Pathish) -> Settings:
        path = Path(path)
        if not path.is_file():
            raise Exception(f"{path} is not a file")

        with path.open("r") as file:
            data = yaml.safe_load(file)

        return cattrs.structure(data, cls)

    @classmethod
    def default(cls) -> Settings:
        return cls(
            generator_temps=[0.7, 0.7, 0.7],
            researcher_temp=0.5,
            resolver_temp=0.5,
            model="gpt-4",
            mode=Mode.RESOLVER,
            credentials=Credentials.dummy(),
            verbosity=Verbosity.SOME,
            vi_mode=False,
        )


class Verbosity(strenum.StrEnum):
    NONE = strenum.auto()
    SOME = strenum.auto()
    ALL = strenum.auto()


DUMMY_KEY = "XXXXXX"


def _api_key_repr(key: str) -> str:
    if key == DUMMY_KEY:
        return key

    assert len(key) > 8
    return key[:4] + "*" * 4 + key[-4:]


@attrs.define
class Credentials:
    key: str = attrs.field(repr=_api_key_repr)

    @classmethod
    def dummy(cls) -> Credentials:
        return cls(DUMMY_KEY)


class Mode(strenum.StrEnum):
    """Enum representing the different modes of interaction."""

    ZERO_SHOT = strenum.auto()
    STEP_BY_STEP = strenum.auto()
    RESOLVER = strenum.auto()


@attrs.define(frozen=True)
class GPTConfig:
    """Configuration for the GPT model.

    Attributes:
        model:
            The GPT model to use (default is 'gpt-4').
        credentials:
            Credentials for accessing the model.
        mode:
            The mode of interaction.
    """

    model: str = attrs.field(default="gpt-4")
    credentials: Credentials = attrs.field(default=Settings.default().credentials)
    mode: Mode = attrs.field(default=Mode.ZERO_SHOT)

    @classmethod
    def default(cls) -> GPTConfig:
        """Factory method to create a default GPTConfig.

        Returns:
            GPTConfig:
                A GPTConfig instance with default values.
        """
        return cls()


class Role(strenum.StrEnum):
    """Enum representing the different roles in a conversation"""

    USER = strenum.auto()
    ASSISTANT = strenum.auto()
    SYSTEM = strenum.auto()


@attrs.define(frozen=True)
class Message:
    """Structure for encapsulating a message in a conversation.

    Attributes:
        role:
            The role of the sender of the message.
        content:
            The content of the message.
    """

    role: Role
    content: str


@attrs.define
class Response:
    """Structure for encapsulating a response from the GPT model.

    Attributes:
        message:
            The message from the model.
        total_tokens:
            The total number of tokens used.
        finish_reason:
            The reason for finishing the response.
    """

    message: Message
    total_tokens: int
    finish_reason: str

    @classmethod
    def from_openai_response(cls, response) -> Response:
        """Factory method to create a Response from an OpenAI API response.

        Args:
            response:
                The response from the OpenAI API.

        Returns:
            Response:
                A Response instance with values from the API response.
        """
        return cls(
            message=cattrs.structure(response["choices"][0]["message"], Message),
            total_tokens=response["usage"]["total_tokens"],
            finish_reason=response["choices"][0]["finish_reason"],
        )



"""This module is a wrapper around OpenAI's GPT model

It defines two main classes, Agent
and SmartGPT, which together facilitate interactive sessions with the AI model. The
module supports different modes of interaction, including zero-shot, "step-by-step"
chain of thought prompt flavoring, and the full "SmartGPT" resolver mode.

Classes:
    Agent:
        Represents an interface to the GPT model.
    SmartGPT:
        Manages interactive sessions with the AI, supporting different modes of
        interaction.
"""

from __future__ import annotations

import copy
import logging
import time
from typing import Dict, List, Optional

import attrs
import openai
from openai.error import InvalidRequestError, RateLimitError

from smartgpt import prompts
from smartgpt.datatypes import (
    Credentials,
    GPTConfig,
    Message,
    Mode,
    Response,
    Role,
    Settings,
    Verbosity,
)
from smartgpt.logger import default_logger, get_logger
from smartgpt.repl import repl
from smartgpt.user_profile import SETTINGS_PATH


@attrs.define
class Agent:
    """Represents an interface to the GPT model.

    It encapsulates the process of sending messages to the model and receiving
    responses. The class also handles message history.

    Attributes:
        messages:
            A list of messages sent to and received from the model.
        credentials:
            Credentials for accessing the model.
        model:
            The model to use (default is 'gpt-4').
        temp:
            The temperature parameter to use when generating responses.
    """

    messages: List[Dict[str, str]] = attrs.field(factory=list)
    credentials: Credentials = attrs.field(default=Settings.default().credentials)
    model: str = attrs.field(default="gpt-4")
    temp: float = attrs.field(default=0.5)

    def append_message(self, message: Message) -> None:
        """Appends a message to the current message history.

        Args:
            message: The message to append.
        """
        self.messages.append(attrs.asdict(message))

    def request(self) -> Response:
        """Sends the current message history to the GPT model via an API request

        The message history includes all previous interactions, which allows the model
        to generate a response based on the entire conversation context.

        Returns:
            Response:
                A Response object that encapsulates the model's response, which includes
                the generated message, remaining tokens, and other metadata.
        """

        try:
            return Response.from_openai_response(
                openai.ChatCompletion.create(
                    model=self.model,
                    messages=self.messages,
                    api_key=self.credentials.key,
                )
            )
        except RateLimitError:
            default_logger.info("Hit rate limit. Sleeping for 20 seconds...")
            time.sleep(20)
            return self.request()
        except InvalidRequestError:
            raise NotImplementedError()

    def response(self, prompt: str) -> Message:
        """Appends prompt to message history and sends request to the GPT model.

        The model's response is then appended to the message history.

        Args:
            prompt:
                The prompt to send to the model.

        Returns:
            Message:
                The model's response encapsulated in a Message object.
        """
        self.append_message(Message(Role.USER, prompt))

        response = self.request()

        self.append_message(response.message)

        return response.message


@attrs.define
class SmartGPT:
    """Manages interactive sessions with the AI, supports different interaction modes.

    In zero-shot mode, the user's prompts are sent directly to the model. In
    step-by-step mode, prompts are flavored with the "step-by-step" chain of thought
    prompt. In resolver mode, the "SmartGPT" mode is used, where multiple agents
    generate responses, which are scrutinized by a researcher agent and then a final
    response is chosen by a resolver agent.

    Attributes:
        config:
            The configuration for the interaction.
        main:
            The main agent, responsible for sending prompts, receiving responses, and
            most importantly, storing a linear chat history.
        researcher:
            An agent responsible for scrutinizing generator responses.
        resolver:
            An agent responsible for considering the researcher's analysis and creating
            the final response.
        generators:
            A list of agents for generating responses in resolver mode.
    """

    config: GPTConfig
    main: Agent
    researcher: Agent
    resolver: Agent
    generators: List[Agent]
    verbosity: Verbosity = Verbosity.SOME

    logger: logging.Logger = attrs.field(init=False)

    def __attrs_post_init__(self) -> None:
        self.logger = get_logger(verbosity=self.verbosity)

    def create_response(self, prompt: str) -> Message:
        """
        Determines the mode of interaction based on the current configuration and
        generates a response from the model accordingly.

        In ZERO_SHOT mode, the prompt is sent directly to the main agent, which sends it
        to the model and returns the response.

        In STEP_BY_STEP mode, the prompt is transformed using the `prompts.step_by_step`
        function before being sent to the model. The transformed prompt is intended to
        guide the model to generate a more detailed response.

        In RESOLVER mode, multiple generators create potential responses to the prompt.
        The researcher agent points out the flaws in each generated response. Finally,
        the resolver agent combines the selected responses into a final response.

        Args:
            prompt:
                The prompt to send to the model.

        Returns:
            Message:
                The model's response encapsulated in a Message object.
        """

        self.logger.debug(f"Creating response for prompt: '{prompt}'")

        if self.config.mode == Mode.ZERO_SHOT:
            self.logger.info("Generating response for zero-shot prompt...")
            self.logger.debug("Mode: ZERO_SHOT")

            response = self.main.response(prompt)
            self.logger.debug(f"ZERO_SHOT response:\n{response.content}")

        elif self.config.mode == Mode.STEP_BY_STEP:
            self.logger.info(
                "Generating response for chain-of-thought-flavored prompt..."
            )
            self.logger.debug("Mode: STEP_BY_STEP")

            transformed_prompt = prompts.step_by_step(prompt)
            self.logger.debug(f"Transformed prompt:\n{transformed_prompt}")

            response = self.main.response(transformed_prompt)
            self.logger.debug(f"STEP_BY_STEP response:\n{response.content}")

        elif self.config.mode == Mode.RESOLVER:
            self.logger.debug("Mode: RESOLVER")

            candidates: List[str] = []
            candidate_prompt = prompts.step_by_step(prompt)
            self.logger.debug(f"Prompt for generators:\n{candidate_prompt}")

            for i, generator in enumerate(self.generators):
                self.logger.info(f"Generating response for generator {i+1}...")

                generator.messages = copy.deepcopy(self.main.messages)

                candidate = generator.response(candidate_prompt).content
                candidates.append(candidate)
                self.logger.debug(f"Generator {i+1} response:\n{candidate}")

            self.logger.info("Generating response for researcher...")

            researcher_prompt = prompts.you_are_a_researcher(prompt, candidates)
            self.logger.debug(f"Researcher prompt:\n{researcher_prompt}")

            self.researcher.messages = copy.deepcopy(self.main.messages)

            researcher_response = self.researcher.response(researcher_prompt)
            self.logger.debug(f"Researcher response:\n{researcher_response.content}")

            self.logger.info("Generating response for resolver...")

            resolver_prompt = prompts.you_are_a_resolver(
                candidates, silence_rationale=(self.verbosity != Verbosity.ALL)
            )
            self.logger.debug(f"Resolver prompt:\n{resolver_prompt}")

            self.resolver.messages = copy.deepcopy(self.researcher.messages)
            response = self.resolver.response(resolver_prompt)

            # Add the prompt and response to the main agent
            self.main.append_message(Message(Role.USER, candidate_prompt))
            self.main.append_message(response)

        logging.debug("Finished generating response")

        return response

    @classmethod
    def create(cls, settings: Optional[Settings] = None) -> SmartGPT:
        if settings is None:
            settings = Settings.load(SETTINGS_PATH)

        return cls(
            config=GPTConfig(
                credentials=settings.credentials,
                model=settings.model,
                mode=settings.mode,
            ),
            main=Agent(
                credentials=settings.credentials,
                model=settings.model,
                temp=settings.resolver_temp,
            ),
            researcher=Agent(
                credentials=settings.credentials,
                model=settings.model,
                temp=settings.researcher_temp,
            ),
            resolver=Agent(
                credentials=settings.credentials,
                model=settings.model,
                temp=settings.resolver_temp,
            ),
            generators=[
                Agent(
                    credentials=settings.credentials,
                    model=settings.model,
                    temp=settings.generator_temps[i],
                )
                for i in range(settings.num_agents)
            ],
            verbosity=settings.verbosity,
        )

    def repl(self) -> None:
        repl(self)



#! /usr/bin/env python

import sys
from setuptools import find_packages, setup

if sys.version_info < (3, 8):
    sys.exit("\nERROR: Python versions lower than 3.8 are not supported\n")

setup(
    name="smartgpt",
    version="0.1.2",
    packages=find_packages(),
    scripts=["bin/smartgpt"],
    author_email="kiefl.evan@gmail.com",
    author="Evan Kiefl",
    url="https://github.com/ekiefl/smartgpt",
    install_requires=[
        "openai",
        "attrs",
        "cattrs",
        "pyyaml",
        "prompt_toolkit",
    ],
    include_package_data=True,
    zip_safe=False,
)



from typing import Protocol, Tuple

from prompt_toolkit import PromptSession
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.history import FileHistory

from smartgpt.datatypes import Message, Settings
from smartgpt.user_profile import MAX_HISTORY, REPL_HISTORY_PATH, SETTINGS_PATH


class LimitedFileHistory(FileHistory):
    def __init__(self, filename, max_size):
        super().__init__(filename)
        self.max_size = max_size

    def append_string(self, string):
        super().append_string(string)
        self.limit_file_size()

    def limit_file_size(self):
        with open(self.filename, "r+") as f:
            lines = f.readlines()
            if len(lines) > self.max_size:
                f.seek(0)
                f.truncate()
                f.writelines(lines[-self.max_size :])  # Keep last max_size lines


session: PromptSession = PromptSession(
    history=LimitedFileHistory(str(REPL_HISTORY_PATH), MAX_HISTORY),
    multiline=True,
    wrap_lines=True,
    vi_mode=Settings.load(SETTINGS_PATH).vi_mode,
)


USER_PREFIX = FormattedText(
    [
        ("", "(To submit: press ESC, then ENTER)\n"),
        ("bold fg:ansigreen", "> "),
    ]
)


class Model(Protocol):
    def create_response(self, prompt: str) -> Message:
        ...


def repl(model: Model) -> None:
    """Starts the Read-Eval-Print-Loop (REPL) for interaction with a model

    This function runs indefinitely until the session is manually terminated.
    """
    while True:
        _, _ = repl_iteration(model)


def repl_iteration(model: Model, quiet=False) -> Tuple[str, str]:
    """Executes a single pass of the Read-Eval-Print-Loop (REPL)

    Args:
        quiet:
            If true, the function will not print the model's response. Default is
            False.

    Returns:
        Tuple[str, str]:
            The user's prompt and the model's response.
    """
    prompt = session.prompt(USER_PREFIX)

    response_message = model.create_response(prompt)

    if not quiet:
        print(f"\n{response_message.content}\n")

    return prompt, response_message.content



import sys
from pathlib import Path

from smartgpt.datatypes import Credentials, Settings
from smartgpt.error import SmartGPTError

SETTINGS_DIR = Path.home() / ".smartgpt"
SETTINGS_DIR.mkdir(exist_ok=True)

SETTINGS_PATH = SETTINGS_DIR / "settings.yaml"
if not SETTINGS_PATH.exists():
    Settings.default().save(SETTINGS_PATH)

settings = Settings.load(SETTINGS_PATH)
if settings.credentials == (dummy := Credentials.dummy()):
    print(
        SmartGPTError(
            f"Almost! In order to use Smart-GPT, you need your OpenAI API key. Get "
            f"yours at https://platform.openai.com/account/api-keys, then open "
            f"'{SETTINGS_PATH}' and replace '{dummy.key}' with your API key."
        )
    )
    sys.exit(1)

REPL_HISTORY_PATH = SETTINGS_DIR / "history.txt"
MAX_HISTORY = 3000
